"""Unit tests for a Newport single axis motion controller."""
from typing import cast
import unittest
from unittest.mock import Mock, call, patch

import qmi
from qmi.core.exceptions import QMI_InstrumentException
from qmi.core.transport import QMI_TcpTransport
from qmi.instruments.newport import Newport_Smc100Cc, ControlLoopState
from qmi.instruments.newport.actuators import TRA12CC, TRB6CC, CMA25CCL


class TestNewportSmc100Cc(unittest.TestCase):
    """
    Tests for the single axis motion controller.
    """

    TRANSPORT_STR = "/dev/cu.usbserial-FT5TMFGL"

    def setUp(self):
        qmi.start("Test100CcControllerContext", console_loglevel="CRITICAL")
        # Add patches
        patcher = patch(
            'qmi.instruments.newport.single_axis_motion_controller.create_transport', spec=QMI_TcpTransport)
        self._transport_mock: Mock = patcher.start().return_value
        self.addCleanup(patcher.stop)
        patcher = patch(
            'qmi.instruments.newport.single_axis_motion_controller.ScpiProtocol', autospec=True)
        self._scpi_mock: Mock = patcher.start().return_value
        self.addCleanup(patcher.stop)
        # Make DUTs
        self.controller_address = 1
        Newport_Smc100Cc.PW0_EXEC_TIME = 0.1
        Newport_Smc100Cc.COMMAND_EXEC_TIME = 0.01
        self.instr: Newport_Smc100Cc = qmi.make_instrument(
            "sam_controller", Newport_Smc100Cc, self.TRANSPORT_STR, "FT5TMFGL",
            {1: TRA12CC, 2: TRB6CC, 3: CMA25CCL},
            90210)
        self.instr = cast(Newport_Smc100Cc, self.instr)
        self.instr.open()

    def tearDown(self):
        self.instr.close()
        qmi.stop()

    def test_set_encoder_increment_sets_increment(self):
        """Test set encoder increment."""
        state = "14"  # Set to CONFIGURATION state to go to _enter_configuration_state.
        increment = 0.004
        self._scpi_mock.ask.side_effect = [f"1TS0000{state}", "@"]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}TE\r\n")
        ]
        expected_write_calls = [
            call(f"{self.controller_address}SU{increment}\r\n"),
            call(f"{self.controller_address}PW0\r\n")
        ]

        self.instr.set_encoder_increment_value(increment)

        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_encoder_increment_with_controller_address_sets_increment(self):
        """Test set encoder increment with controller address."""
        state = "14"  # Set to CONFIGURATION state to go to _enter_configuration_state.
        increment = 0.004
        controller_address = 3
        self._scpi_mock.ask.side_effect = [f"1TS0000{state}", "@"]
        expected_ask_calls = [
            call(f"{controller_address}TS\r\n"),
            call(f"{controller_address}TE\r\n")
        ]
        expected_write_calls = [
            call(f"{controller_address}SU{increment}\r\n"),
            call(f"{controller_address}PW0\r\n")
        ]

        self.instr.set_encoder_increment_value(increment, controller_address)

        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_get_encoder_increment_value_gets_value(self):
        """Test get encoder increment value."""
        state = "14"  # Set to CONFIGURATION state to go to _enter_configuration_state.
        encoder_resolution = TRA12CC.ENCODER_RESOLUTION
        expected_encoder_unit = 10
        self._scpi_mock.ask.side_effect = [
            f"1TS0000{state}", f"{self.controller_address}SU %s" % (encoder_resolution / expected_encoder_unit), "@"]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}SU?\r\n"),
            call(f"{self.controller_address}TE\r\n")
        ]
        expected_write_calls = [
            call(f"{self.controller_address}PW0\r\n"),
        ]

        actual_encoder_unit = self.instr.get_encoder_increment_value()

        self.assertEqual(expected_encoder_unit, actual_encoder_unit)
        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_get_driver_voltage(self):
        """Test getting driver voltage."""
        state = "14"  # Set to CONFIGURATION state to go to _enter_configuration_state.
        expected_value = 24
        self._scpi_mock.ask.side_effect = [
            f"1TS0000{state}", f"{self.controller_address}DV %i" % expected_value, "@"]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}DV?\r\n"),
            call(f"{self.controller_address}TE\r\n")
        ]
        expected_write_calls = [
            call(f"{self.controller_address}PW0\r\n"),
        ]

        actual_value = self.instr.get_driver_voltage()

        self.assertEqual(actual_value, expected_value)
        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_driver_voltage(self):
        """Test setting driver voltage."""
        state = "14"  # Set to CONFIGURATION state to go to _enter_configuration_state.
        driver_voltage = 24
        self._scpi_mock.ask.side_effect = [f"1TS0000{state}", "@"]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}TE\r\n")
        ]
        expected_write_calls = [
            call(f"1DV{driver_voltage}\r\n"),
            call(f"{self.controller_address}PW0\r\n")
        ]

        self.instr.set_driver_voltage(driver_voltage)
        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_driver_voltage_excepts(self):
        """Test setting driver voltage excepts when value out of valid range."""
        driver_voltage_invalid = [11.9, 48.1]
        for value in driver_voltage_invalid:
            with self.assertRaises(QMI_InstrumentException):
                self.instr.set_driver_voltage(value)

        self._scpi_mock.write.assert_not_called()
        self._scpi_mock.ask.assert_not_called()

    def test_get_low_pass_filter_cutoff_frequency_gets_frequency(self):
        """Test get low pass filter cutoff frequency."""
        state = "3C"  # Set to DISABLE state to get the shortest path.
        expected_frequency = 1.0
        self._scpi_mock.ask.side_effect = [
            f"1TS0000{state}", f"{self.controller_address}FD%s" % expected_frequency, "@"]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}FD?\r\n"),
            call(f"{self.controller_address}TE\r\n")
        ]
        expected_write_calls = [
            call(f"{self.controller_address}MM0\r\n"),
        ]

        actual_frequency = self.instr.get_low_pass_filter_cutoff_frequency()

        self.assertEqual(expected_frequency, actual_frequency)
        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_get_low_pass_filter_cutoff_frequency_gets_frequency_in_configuration(self):
        """Test get low pass filter cutoff frequency in CONFIGURATION state."""
        state = "14"  # Set to CONFIGURATION state to exit through config_state exit.
        expected_frequency = 1.0
        self._scpi_mock.ask.side_effect = [
            f"1TS0000{state}", f"1TS0000{state}", f"{self.controller_address}FD%s" % expected_frequency, "@"]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}FD?\r\n"),
            call(f"{self.controller_address}TE\r\n")
        ]
        expected_write_calls = [
            call(f"{self.controller_address}PW0\r\n"),
        ]

        actual_frequency = self.instr.get_low_pass_filter_cutoff_frequency()

        self.assertEqual(expected_frequency, actual_frequency)
        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_low_pass_filter_cutoff_frequency_with_persist_sets_frequency(self):
        """Test set low pass filter cutoff frequency with persist."""
        state = "14"  # Set to CONFIGURATION state to get the shortest path.
        frequency = 1.3
        persist = True
        self._scpi_mock.ask.side_effect = [f"1TS0000{state}", "@"]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}TE\r\n")
        ]
        expected_write_calls = [
            call(f"{self.controller_address}FD{frequency}\r\n"),
            call(f"{self.controller_address}PW0\r\n"),
        ]

        self.instr.set_low_pass_filter_cutoff_frequency(frequency, persist)

        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_low_pass_filter_cutoff_frequency_sets_frequency(self):
        """Test set low pass filter cutoff frequency."""
        state = "3C"  # Set to DISABLE state to get the shortest path.
        frequency = 1.3
        self._scpi_mock.ask.side_effect = [f"1TS0000{state}", "@"]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}TE\r\n")
        ]
        expected_write_calls = [
            call(f"{self.controller_address}FD{frequency}\r\n"),
            call(f"{self.controller_address}MM0\r\n"),
        ]

        self.instr.set_low_pass_filter_cutoff_frequency(frequency)

        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_low_pass_filter_cutoff_frequency_with_error_throws_exception(self):
        """Test set low pass filter cutoff frequency with error."""
        out_of_bounds = [1E-6, 2000]
        for value in out_of_bounds:
            with self.assertRaises(QMI_InstrumentException):
                self.instr.set_low_pass_filter_cutoff_frequency(value)

        self._scpi_mock.write.assert_not_called()
        self._scpi_mock.ask.assert_not_called()

    def test_get_following_error_limit_gets_limit(self):
        """Test get following error limit."""
        state = "3C"  # Set to DISABLE state to get the shortest path.
        expected_limit = 1.0
        self._scpi_mock.ask.side_effect = [
            f"1TS0000{state}", f"{self.controller_address}FE%s" % expected_limit, "@"]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}FE?\r\n"),
            call(f"{self.controller_address}TE\r\n")
        ]
        expected_write_calls = [
            call(f"{self.controller_address}MM0\r\n"),
        ]

        actual_limit = self.instr.get_following_error_limit()

        self.assertEqual(expected_limit, actual_limit)
        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_get_following_error_limit_gets_limit_in_configuration(self):
        """Test get following error limit gets limit in CONFIGURATION state."""
        state = "14"  # Set to CONFIGURATION state to exit through config_state exit.
        expected_limit = 1.0
        self._scpi_mock.ask.side_effect = [
            f"1TS0000{state}", f"1TS0000{state}", f"{self.controller_address}FE%s" % expected_limit, "@"]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}FE?\r\n"),
            call(f"{self.controller_address}TE\r\n")
        ]
        expected_write_calls = [
            call(f"{self.controller_address}PW0\r\n"),
        ]

        actual_limit = self.instr.get_following_error_limit()

        self.assertEqual(expected_limit, actual_limit)
        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_following_error_limit_with_persist_sets_limit(self):
        """Test set following error limit with persist."""
        state = "14"  # Set to CONFIGURATION state to get the shortest path.
        limit = 1.3
        persist = True
        self._scpi_mock.ask.side_effect = [f"1TS0000{state}", "@"]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}TE\r\n")
        ]
        expected_write_calls = [
            call(f"{self.controller_address}FE{limit}\r\n"),
            call(f"{self.controller_address}PW0\r\n"),
        ]

        self.instr.set_following_error_limit(limit, persist)

        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_following_error_limit_sets_limit(self):
        """Test set following error limit."""
        state = "3C"  # Set to DISABLE state to get the shortest path.
        limit = 1.3
        self._scpi_mock.ask.side_effect = [f"1TS0000{state}", "@"]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}TE\r\n")
        ]
        expected_write_calls = [
            call(f"{self.controller_address}FE{limit}\r\n"),
            call(f"{self.controller_address}MM0\r\n"),
        ]

        self.instr.set_following_error_limit(limit)

        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_following_error_limit_with_error_throws_exception(self):
        """Test set following error limit with error."""
        out_of_bounds = [1E-6, 1E12]
        for value in out_of_bounds:
            with self.assertRaises(QMI_InstrumentException):
                self.instr.set_following_error_limit(value)

        self._scpi_mock.write.assert_not_called()
        self._scpi_mock.ask.assert_not_called()

    def test_get_friction_compensation_gets_compensation(self):
        """Test get friction compensation."""
        state = "3C"  # Set to DISABLE state to get the shortest path.
        expected_compensation = 1.0
        self._scpi_mock.ask.side_effect = [
            f"1TS0000{state}", f"{self.controller_address}FF %s" % expected_compensation, "@"]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}FF?\r\n"),
            call(f"{self.controller_address}TE\r\n")
        ]
        expected_write_calls = [
            call(f"{self.controller_address}MM0\r\n"),
        ]

        actual_compensation = self.instr.get_friction_compensation()

        self.assertEqual(expected_compensation, actual_compensation)
        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_get_friction_compensation_gets_compensation_in_configuration(self):
        """Test get friction compensation gets value in CONFIGURATION state."""
        state = "14"  # Set to CONFIGURATION state to get the config_state exit.
        expected_compensation = 1.0
        self._scpi_mock.ask.side_effect = [
            f"1TS0000{state}", f"1TS0000{state}", f"{self.controller_address}FF %s" % expected_compensation, "@"]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}FF?\r\n"),
            call(f"{self.controller_address}TE\r\n")
        ]
        expected_write_calls = [
            call(f"{self.controller_address}PW0\r\n"),
        ]

        actual_compensation = self.instr.get_friction_compensation()

        self.assertEqual(expected_compensation, actual_compensation)
        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_friction_compensation_with_persist_sets_compensation(self):
        """Test set friction compensation with persist."""
        state = "14"  # Set to CONFIGURATION state.
        driver_voltage = 100.0
        compensation = 1.3
        persist = True
        self._scpi_mock.ask.side_effect = [
            f"1TS0000{state}", f"{self.controller_address}DV {driver_voltage}", "@", "@"
        ]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}DV?\r\n"),
            call(f"{self.controller_address}TE\r\n"),
        ]
        expected_write_calls = [
            call(f"{self.controller_address}FF{compensation}\r\n"),
            call(f"{self.controller_address}PW0\r\n"),
        ]

        self.instr.set_friction_compensation(compensation, persist)

        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_friction_compensation_sets_compensation(self):
        """Test set friction compensation."""
        state = "3C"  # Set to DISABLE state.
        compensation = 1.3
        self._scpi_mock.ask.side_effect = [f"1TS0000{state}", "@"]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}TE\r\n"),
        ]
        expected_write_calls = [
            call(f"{self.controller_address}FF{compensation}\r\n"),
            call(f"{self.controller_address}MM0\r\n"),
        ]

        self.instr.set_friction_compensation(compensation)

        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_friction_compensation_sets_compensation_from_ready(self):
        """Test set friction compensation sets compensation from READY state."""
        state = "32"  # Set to READY state.
        compensation = 1.3
        self._scpi_mock.ask.side_effect = [f"1TS0000{state}", "@", "@"]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}TE\r\n"),
            call(f"{self.controller_address}TE\r\n")
        ]
        expected_write_calls = [
            call(f"{self.controller_address}MM1\r\n"),
            call(f"{self.controller_address}FF{compensation}\r\n"),
            call(f"{self.controller_address}MM0\r\n"),
        ]

        self.instr.set_friction_compensation(compensation)

        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_friction_compensation_throws_exception(self):
        """Test set friction compensation with error on input value being out-of-range."""
        state = "14"  # Set to CONFIGURATION state directly.
        persist = True  # We can do this only in persist mode
        expected_exceptions = [
            "Provided value %.3f not in valid range 0 <= friction_compensation < %.1f.",
            "Provided value %.1f not in valid range 0 <= friction_compensation < %.1f."
        ]
        driver_voltage = 1.0
        self._scpi_mock.ask.side_effect = [
            f"1TS0000{state}", f"{self.controller_address}DV {driver_voltage}", "@",
            f"1TS0000{state}", f"{self.controller_address}DV {driver_voltage}", "@"
        ]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}DV?\r\n"),
            call(f"{self.controller_address}TE\r\n"),
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}DV?\r\n"),
            call(f"{self.controller_address}TE\r\n")
        ]
        expected_write_calls = [
            call(f"{self.controller_address}PW0\r\n"),
            call(f"{self.controller_address}PW0\r\n")
        ]

        out_of_bounds = [-0.001, driver_voltage]
        for e, value in enumerate(out_of_bounds):
            with self.assertRaises(QMI_InstrumentException) as exc:
                self.instr.set_friction_compensation(value, persist)

            self.assertEqual(expected_exceptions[e] % (value, driver_voltage), str(exc.exception))

        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_get_derivative_gain_gets_gain(self):
        """Test get derivative gain."""
        state = "3C"  # Set to DISABLE state to get the shortest path.
        expected_gain = 1.0
        self._scpi_mock.ask.side_effect = [
            f"1TS0000{state}", f"{self.controller_address}KD%s" % expected_gain, "@"]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}KD?\r\n"),
            call(f"{self.controller_address}TE\r\n")
        ]
        expected_write_calls = [
            call(f"{self.controller_address}MM0\r\n"),
        ]

        actual_gain = self.instr.get_derivative_gain()

        self.assertEqual(expected_gain, actual_gain)
        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_derivative_gain_with_persist_sets_gain(self):
        """Test set derivative gain with persist."""
        state = "14"  # Set to CONFIGRATION state.
        gain = 1.3
        persist = True
        self._scpi_mock.ask.side_effect = [f"1TS0000{state}", "@"]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}TE\r\n")
        ]
        expected_write_calls = [
            call(f"{self.controller_address}KD%s\r\n" % gain),
            call(f"{self.controller_address}PW0\r\n"),
        ]

        self.instr.set_derivative_gain(gain, persist)

        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_derivative_gain_sets_gain(self):
        """Test set derivative gain."""
        state = "3C"  # Set to DISABLE state to get the shortest path.
        gain = 1.3
        self._scpi_mock.ask.side_effect = [f"1TS0000{state}", "@"]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}TE\r\n")
        ]
        expected_write_calls = [
            call(f"{self.controller_address}KD{gain}\r\n"),
            call(f"{self.controller_address}MM0\r\n"),
        ]

        self.instr.set_derivative_gain(gain)

        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_derivative_gain_sets_gain_from_ready(self):
        """Test set derivative gain sets gain also from READY state."""
        state = "34"  # Set to READY state.
        gain = 1.3
        self._scpi_mock.ask.side_effect = [f"1TS0000{state}", "@", "@"]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}TE\r\n"),
            call(f"{self.controller_address}TE\r\n")
        ]
        expected_write_calls = [
            call(f"{self.controller_address}MM1\r\n"),
            call(f"{self.controller_address}KD{gain}\r\n"),
            call(f"{self.controller_address}MM0\r\n"),
        ]

        self.instr.set_derivative_gain(gain)

        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_derivative_gain_with_error_throws_exception(self):
        """Test set derivative gain with error."""
        out_of_bounds = [-1E-6, 1E12]
        for value in out_of_bounds:
            with self.assertRaises(QMI_InstrumentException):
                self.instr.set_derivative_gain(value)

        self._scpi_mock.write.assert_not_called()
        self._scpi_mock.ask.assert_not_called()

    def test_get_integral_gain_gets_gain(self):
        """Test get integral gain."""
        state = "3C"  # Set to DISABLE state to get the shortest path.
        expected_gain = 1.0
        self._scpi_mock.ask.side_effect = [
            f"1TS0000{state}", f"{self.controller_address}KI %s" % expected_gain, "@"]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}KI?\r\n"),
            call(f"{self.controller_address}TE\r\n")
        ]
        expected_write_calls = [
            call(f"{self.controller_address}MM0\r\n"),
        ]

        actual_gain = self.instr.get_integral_gain()

        self.assertEqual(expected_gain, actual_gain)
        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_get_integral_gain_gets_gain_in_configuration(self):
        """Test get integral gain gets gain in CONFIGURATION state."""
        state = "14"  # Set to CONFIGURATION state to exit through config_state exit.
        expected_gain = 1.0
        self._scpi_mock.ask.side_effect = [
            f"1TS0000{state}", f"1TS0000{state}", f"{self.controller_address}KI %s" % expected_gain, "@"]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}KI?\r\n"),
            call(f"{self.controller_address}TE\r\n")
        ]
        expected_write_calls = [
            call(f"{self.controller_address}PW0\r\n"),
        ]

        actual_gain = self.instr.get_integral_gain()

        self.assertEqual(expected_gain, actual_gain)
        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_integral_gain_with_persist_sets_gain(self):
        """Test set integral gain with persist."""
        state = "14"  # Set to CONFIGURATION state.
        gain = 1.3
        persist = True
        self._scpi_mock.ask.side_effect = [f"1TS0000{state}", "@"]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}TE\r\n")
        ]
        expected_write_calls = [
            call(f"{self.controller_address}KI%s\r\n" % gain),
            call(f"{self.controller_address}PW0\r\n"),
        ]

        self.instr.set_integral_gain(gain, persist)

        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_integral_gain_sets_gain(self):
        """Test set integral gain."""
        state = "3C"  # Set to DISABLE state.
        gain = 1.3
        self._scpi_mock.ask.side_effect = [f"1TS0000{state}", "@"]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}TE\r\n")
        ]
        expected_write_calls = [
            call(f"{self.controller_address}KI%s\r\n" % gain),
            call(f"{self.controller_address}MM0\r\n"),
        ]

        self.instr.set_integral_gain(gain)

        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_integral_gain_with_error_throws_exception(self):
        """Test set integral gain with error."""
        out_of_bounds = [-1E-6, 1E12]
        for value in out_of_bounds:
            with self.assertRaises(QMI_InstrumentException):
                self.instr.set_integral_gain(value)

        self._scpi_mock.write.assert_not_called()
        self._scpi_mock.ask.assert_not_called()

    def test_get_proportional_gain_gets_gain(self):
        """Test get proportional gain."""
        state = "3C"  # Set to DISABLE state to get the shortest path.
        expected_gain = 1.0
        self._scpi_mock.ask.side_effect = [
            f"1TS0000{state}", f"{self.controller_address}KP %s" % expected_gain, "@"]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}KP?\r\n"),
            call(f"{self.controller_address}TE\r\n")
        ]
        expected_write_calls = [
            call(f"{self.controller_address}MM0\r\n"),
        ]

        actual_gain = self.instr.get_proportional_gain()

        self.assertEqual(expected_gain, actual_gain)
        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_get_proportional_gain_gets_gain_in_configuration(self):
        """Test get proportional gain in CONFIGURATION state."""
        state = "14"  # Set to CONFIGURATION state.
        expected_gain = 1.0
        self._scpi_mock.ask.side_effect = [
            f"1TS0000{state}", f"1TS0000{state}", f"{self.controller_address}KP %s" % expected_gain, "@"]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}KP?\r\n"),
            call(f"{self.controller_address}TE\r\n")
        ]
        expected_write_calls = [
            call(f"{self.controller_address}PW0\r\n"),
        ]

        actual_gain = self.instr.get_proportional_gain()

        self.assertEqual(expected_gain, actual_gain)
        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_proportional_gain_with_persist_sets_gain(self):
        """Test set proportional gain with persist."""
        state = "14"  # Set to CONFIGURATION state.
        gain = 1.3
        persist = True
        self._scpi_mock.ask.side_effect = [f"1TS0000{state}", "@"]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}TE\r\n")
        ]
        expected_write_calls = [
            call(f"{self.controller_address}KP{gain}\r\n"),
            call(f"{self.controller_address}PW0\r\n"),
        ]

        self.instr.set_proportional_gain(gain, persist)

        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_proportional_gain_sets_gain(self):
        """Test set proportional gain."""
        state = "3C"  # Set to DISABLE state.
        gain = 1.3
        self._scpi_mock.ask.side_effect = [f"1TS0000{state}", "@"]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}TE\r\n")
        ]
        expected_write_calls = [
            call(f"{self.controller_address}KP{gain}\r\n"),
            call(f"{self.controller_address}MM0\r\n"),
        ]

        self.instr.set_proportional_gain(gain)

        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_proportional_gain_with_error_throws_exception(self):
        """Test set proportional gain with error."""
        out_of_bounds = [-1E-6, 1E12]
        for value in out_of_bounds:
            with self.assertRaises(QMI_InstrumentException):
                self.instr.set_proportional_gain(value)

        self._scpi_mock.write.assert_not_called()
        self._scpi_mock.ask.assert_not_called()

    def test_get_velocity_feed_forward_gets_vff(self):
        """Test get velocity feed forward."""
        state = "3C"  # Set to DISABLE state to get the shortest path.
        expected_vff = 1.0
        self._scpi_mock.ask.side_effect = [
            f"1TS0000{state}", f"{self.controller_address}KV%s" % expected_vff, "@"]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}KV?\r\n"),
            call(f"{self.controller_address}TE\r\n")
        ]
        expected_write_calls = [
            call(f"{self.controller_address}MM0\r\n"),
        ]

        actual_vff = self.instr.get_velocity_feed_forward()

        self.assertEqual(expected_vff, actual_vff)
        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_get_velocity_feed_forward_gets_vff_in_configuration(self):
        """Test get velocity feed forward in CONFIGURATION state."""
        state = "14"  # Set to CONFIGURATION state.
        expected_vff = 1.0
        self._scpi_mock.ask.side_effect = [
            f"1TS0000{state}", f"1TS0000{state}", f"{self.controller_address}KV%s" % expected_vff, "@"]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}KV?\r\n"),
            call(f"{self.controller_address}TE\r\n")
        ]
        expected_write_calls = [
            call(f"{self.controller_address}PW0\r\n"),
        ]

        actual_vff = self.instr.get_velocity_feed_forward()

        self.assertEqual(expected_vff, actual_vff)
        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_velocity_feed_forward_with_persist_sets_vff(self):
        """Test set velocity feed forward with persist."""
        state = "14"  # Set to CONFIGURATION state.
        vff = 1.3
        persist = True
        self._scpi_mock.ask.side_effect = [f"1TS0000{state}", "@"]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}TE\r\n")
        ]
        expected_write_calls = [
            call(f"{self.controller_address}KV%s\r\n" % vff),
            call(f"{self.controller_address}PW0\r\n"),
        ]

        self.instr.set_velocity_feed_forward(vff, persist)

        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_velocity_feed_forward_sets_vff(self):
        """Test set velocity feed forward."""
        state = "3C"  # Set to DISABLE state.
        vff = 1.3
        persist = True
        self._scpi_mock.ask.side_effect = [f"1TS0000{state}", "@"]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}TE\r\n")
        ]
        expected_write_calls = [
            call(f"{self.controller_address}KV%s\r\n" % vff),
            call(f"{self.controller_address}MM0\r\n"),
        ]

        self.instr.set_velocity_feed_forward(vff)

        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_velocity_feed_forward_with_error_throws_exception(self):
        """Test set velocity feed forward with error."""
        out_of_bounds = [-1E-6, 1E12]
        for value in out_of_bounds:
            with self.assertRaises(QMI_InstrumentException):
                self.instr.set_velocity_feed_forward(value)

        self._scpi_mock.write.assert_not_called()
        self._scpi_mock.ask.assert_not_called()

    def test_get_control_loop_state(self):
        """Test getting control loop state."""
        state = "3C"  # Set to DISABLE state to get the shortest path.
        expected_state = 1
        self._scpi_mock.ask.side_effect = [
            f"1TS0000{state}", f"{self.controller_address}SC %s" % expected_state, "@"]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}SC?\r\n"),
            call(f"{self.controller_address}TE\r\n")
        ]
        expected_write_calls = [
            call(f"{self.controller_address}MM0\r\n"),
        ]

        actual_state = self.instr.get_control_loop_state()
        self.assertEqual(actual_state, ControlLoopState(expected_state))

        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_state_configuration_or_disable_check_from_ready(self):
        """Test getting a controller config state from READY state."""
        state = "34"  # Set to READY state
        response = "0"  # Give a valid response
        self._scpi_mock.ask.side_effect = [
            f"1TS0000{state}", "@", f"{self.controller_address}SC %s" % response, "@"]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}TE\r\n"),
            call(f"{self.controller_address}SC?\r\n"),
            call(f"{self.controller_address}TE\r\n")
        ]
        expected_write_calls = [
            call(f"{self.controller_address}MM1\r\n"),
            call(f"{self.controller_address}MM0\r\n"),
        ]

        actual_state = self.instr.get_control_loop_state()
        self.assertEqual(actual_state, ControlLoopState(int(response)))

        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_state_configuration_or_disable_check_from_configuration(self):
        """Test getting a controller config state from CONFIGURATION state."""
        state = "14"  # Set to CONFIGURATION state to go to _enter_configuration_state.
        expected_state = 1
        self._scpi_mock.ask.side_effect = [
            f"1TS0000{state}", f"1TS0000{state}", f"{self.controller_address}SC %s" % expected_state, "@"]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}SC?\r\n"),
            call(f"{self.controller_address}TE\r\n")
        ]
        expected_write_calls = [
            call(f"{self.controller_address}PW0\r\n"),
        ]

        actual_state = self.instr.get_control_loop_state()
        self.assertEqual(actual_state, ControlLoopState(expected_state))

        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_control_loop_state(self):
        """Test setting control loop state."""
        state = "3C"  # Set to DISABLE state to get the shortest path.
        loop_state = 1
        self._scpi_mock.ask.side_effect = [f"1TS0000{state}", "@"]
        expected_ask_calls = [
            call(f"{self.controller_address}TS\r\n"),
            call(f"{self.controller_address}TE\r\n"),
        ]
        expected_write_calls = [
            call(f"{self.controller_address}SC{loop_state}\r\n"),
            call(f"{self.controller_address}MM0\r\n")
        ]

        self.instr.set_control_loop_state(loop_state)

        self._scpi_mock.write.assert_has_calls(expected_write_calls)
        self._scpi_mock.ask.assert_has_calls(expected_ask_calls)

    def test_set_control_loop_state_excepts(self):
        """Test setting control loop state excepts with type out-of-range."""
        loop_state = 2
        with self.assertRaises(QMI_InstrumentException):
            self.instr.set_control_loop_state(loop_state)

        self._scpi_mock.write.assert_not_called()
        self._scpi_mock.ask.assert_not_called()



"""Instrument driver for the Tenma 72-series power supply units.

Two main model groups in this series are the models similar to 2550 (selected as one base class) and
models similar to 13350 (selected also as a base class). Both groups work with USB-to-serial communication
but the second group has also possibility to use UDP communication. For that IP LAN communication protocol
with related functionalities is present, which can be used either through USB or UDP connection.

Some differences are present, even though the basis is the same, as implemented in the Tenma72_Base class:
- The group based on 13350 model require an EOL character (`\n`, `\r`) to finish the commands while group
  based on 2550 model doesn't.
- The group based on 13350 model does not work when providing the channel number to the serial command, while
  the commands in group based on 2550 model they do.
- For example, a command "ISET1:3.0" works for 2550 group, while "ISET:3.0\n" works for 13350 group.
"""
from dataclasses import dataclass
from enum import Enum
import logging
import re
from time import sleep
from typing import Any, Dict, Optional

from qmi.core.context import QMI_Context
from qmi.core.exceptions import QMI_InstrumentException
from qmi.core.instrument import QMI_Instrument, QMI_InstrumentIdentification
from qmi.core.rpc import rpc_method
from qmi.core.transport import create_transport

# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)


@dataclass
class TenmaChannelMode:
    value: bool
    name: str = "C.C"

    def __post_init__(self):
        if self.value:
            self.name = "C.V"


class TrackingState(Enum):
    INDEPENDENT = 0
    TRACKING_SERIES = 1
    TRACKING_PARALLEL = 2
    UNKNOWN = 99999

    @classmethod
    def _missing_(cls, value):
        return cls.UNKNOWN


class Tenma72_Base(QMI_Instrument):
    """The base class for all Tenma model 72 power supply units.

    Attributes:
        BUFFER_SIZE:      The default buffer size for 'read'
        DEFAULT_BAUDRATE: The default baudrate in case of serial connection is used (Baud).
        MAX_VOLTAGE:      The maximum voltage that can be set with the PSU (Volts).
        MAX_CURRENT:      The maximum current that can be set with the PSU (Amperes).
        SEND_SLEEP_TIME:  A default sleep time for sending data.
        READ_TIMEOUT:     A default timeout for reading data.
    """
    BUFFER_SIZE = 50
    DEFAULT_BAUDRATE = 0
    MAX_VOLTAGE = 0
    MAX_CURRENT = 0
    SEND_SLEEP_TIME = 0.02
    READ_TIMEOUT = 0.2

    def __init__(self, context: QMI_Context, name: str, transport: str) -> None:
        super().__init__(context, name)
        if transport.startswith("serial"):
            self._transport = create_transport(
                transport, default_attributes={"baudrate": self.DEFAULT_BAUDRATE}
            )
        elif transport.startswith("udp"):
            self._transport = create_transport(transport)
        else:
            raise QMI_InstrumentException(f"Transport type {transport} not valid!")

    def _send(self, cmd: str) -> None:
        """Writes a command to the Tenma, just an ascii string

        Parameters:
            cmd: The command to write to the device.
        """
        _logger.info("[%s] Sending message %s", self._name, cmd)
        self._transport.write(bytes(cmd, "ascii"))
        sleep(self.SEND_SLEEP_TIME)  # Sleep some bits to let the message arrive

    def _read(self) -> str:
        """Read the response from the device, try until response received or time-out.

        Returns:
            response: Decoded response string.
        """
        response = self._transport.read_until_timeout(self.BUFFER_SIZE, self.READ_TIMEOUT)
        _logger.info("[%s] Read message %s", self._name, response)
        return response.decode('ascii')

    def _enable_output(self, output: bool, command: str) -> None:
        """Internal function to further communicate the output state to instrument.

        Parameters:
            output:  The expected output state.
            command: The enable command string.
        """
        self._send(command)
        status = self.get_status()
        if status["OutputEnabled"] is not output:
            _logger.exception(
                "[%s] Tried to set output with %s but got %s", self._name, command, status["OutputEnabled"]
            )
            raise QMI_InstrumentException("Setting output failed for an unknown reason")

    @rpc_method
    def open(self) -> None:
        self._check_is_closed()
        _logger.info("Opening connection to %s", self._name)
        self._transport.open()
        super().open()

    @rpc_method
    def close(self) -> None:
        self._check_is_open()
        _logger.info("Closing connection to %s", self._name)
        super().close()
        self._transport.close()

    @rpc_method
    def get_idn(self) -> QMI_InstrumentIdentification:
        """Get instrument identification info.

        Returns:
            QMI_InstrumentIdentification: NamedTuple with vendor, model, serial number and SW version info.
        """
        self._send("*IDN?")
        idn = self._read()
        # Find all relevant info from response. Should be of format TENMA 72-2535 SN:1231345 V2.0
        pattern = re.compile(r'(\b\w+\b[^ ]*|\b72[^ ]*|V[^ ]*|SN:[^ ]*)')

        # Find all matches in the input string
        matches = pattern.findall(idn)

        # Extract the relevant parts from the matches
        vendor = matches[0] if matches else None
        model = next((match for match in matches[1:] if match.startswith('72')), None)
        version = next((match for match in matches[1:] if match.startswith('V')), None)
        serial = next((match for match in matches[1:] if match.startswith('SN:')), None)

        if None in (vendor, model, version, serial):
            _logger.debug("[%s] *IDN? returned %s and matching failed", self._name, idn)
            raise QMI_InstrumentException("(Full) instrument identification failed!")

        assert serial is not None  # For satisfying a mypy check
        return QMI_InstrumentIdentification(
            vendor=vendor,
            model=model,
            serial=serial.strip()[3:],
            version=version
        )

    @rpc_method
    def get_status(self) -> Dict[str, Any]:
        """This method needs to be implemented in the inheriting classes."""
        raise NotImplementedError()

    @rpc_method
    def get_current(self, output_channel: Optional[int] = None) -> float:
        """Get the current setting.

        Parameters:
            output_channel: The channel to send the inquiry to.

        Returns:
            Current setting of the channel in Amperes.
        """
        ch = output_channel or ""
        command = f"ISET{ch}?"
        self._send(command)
        # Response has appended sixth byte from *IDN? to voltage reading due to a firmware bug
        return float(self._read()[:5])

    @rpc_method
    def set_current(self, current: float, output_channel: Optional[int] = None) -> None:
        """Set the current setting.

        Parameters:
            current:        Current for the channel in Amperes.
            output_channel: The channel to send the inquiry to.

        Raises:
            ValueError:     If the current setting is not between 0 and self.MAX_CURRENT.
        """
        if not 0 <= current <= self.MAX_CURRENT:
            raise ValueError(f"Invalid current setting {current}, not between 0 and {self.MAX_CURRENT} A")

        ch = output_channel or ""
        command = f"ISET{ch}:{current:.3f}"
        self._send(command)

    @rpc_method
    def get_voltage(self, output_channel: Optional[int] = None) -> float:
        """Get the voltage setting.

        Parameters:
            output_channel: The channel to send the inquiry to.

        Returns:
            Voltage setting of the channel in Volts.
        """
        ch = output_channel or ""
        command = f"VSET{ch}?"
        self._send(command)
        return float(self._read())

    @rpc_method
    def set_voltage(self, voltage: float, output_channel: Optional[int] = None) -> None:
        """Set the voltage setting.

        Parameters:
            voltage:        Voltage setting for the channel in Volts.
            output_channel: The channel to send the inquiry to.

        Raises:
            ValueError:     If the input voltage setting is not between 0 and self.MAX_VOLTAGE.
        """
        if not 0 <= voltage <= self.MAX_VOLTAGE:
            raise ValueError(f"Invalid current setting {voltage}, not between 0 and {self.MAX_VOLTAGE} V")

        ch = output_channel or ""
        command = f"VSET{ch}:{voltage:.3f}"
        self._send(command)

    @rpc_method
    def enable_output(self, output: bool) -> None:
        """Enable or disable output from the PSU.

        Parameters:
            output: Boolean value to either set output ON (True) or OFF (False).
        """
        raise NotImplementedError()


class Tenma72_2550(Tenma72_Base):
    """Instrument driver for the Tenma 72-2550. The driver is tested with this model, but the respective
    manual is also for models 72-2535, 72-2540, 72-2545, 72-2925, 72-2930, 72-2935, 72-2940 & 72-10480.

    This driver can be used only with (USB-to-)serial communications.

    Attributes:
        DEFAULT_BAUDRATE: The default baudrate in case of serial connection is used (Baud).
        MAX_VOLTAGE:      The maximum voltage that can be set with the PSU (Volts).
        MAX_CURRENT:      The maximum current that can be set with the PSU (Amperes).
    """
    DEFAULT_BAUDRATE = 9600
    MAX_VOLTAGE = 60
    MAX_CURRENT = 3

    @rpc_method
    def get_status(self) -> Dict[str, Any]:
        """Get the power supply status as a dictionary of status values.

        Dictionary composition is:
            Ch1Mode:       "C.V" | "C.C"
            Ch2Mode:       "C.V" | "C.C"
            Tracking:      xx, where xx is byte
                             * 00 = Independent
                             * 01 = Tracking series
                             * 10 = Tracking parallel
            OutputEnabled: True | False
        
        Returns: 
            Dictionary of status values.
        """
        self._send("STATUS?")
        statusByte = ord(self._transport.read(1, 1))  # Read response byte
        ch1mode = bool(statusByte & 0x01)
        ch2mode = bool(statusByte & 0x02)
        tracking = TrackingState((statusByte & 0x0C) >> 2)
        output_enabled = statusByte & 0x40

        return {
            "Ch1Mode": TenmaChannelMode(ch1mode).name,
            "Ch2Mode": TenmaChannelMode(ch2mode).name,
            "Tracking": tracking.name.replace("_", " ").title(),
            "OutputEnabled": bool(output_enabled),
        }

    @rpc_method
    def enable_output(self, output: bool) -> None:
        command = f"OUT{int(output)}"
        self._enable_output(output, command)


class Tenma72_2535(Tenma72_2550):
    """Instrument driver for the Tenma 72-2535, child of 72-2550.

    Attributes:
        MAX_VOLTAGE: The maximum voltage that can be set with the PSU (Volts).
        MAX_CURRENT: The maximum current that can be set with the PSU (Amperes).
    """
    MAX_VOLTAGE = 30
    MAX_CURRENT = 3


class Tenma72_2540(Tenma72_2550):
    """Instrument driver for the Tenma 72-2540, child of 72-2550.

    Attributes:
        MAX_VOLTAGE: The maximum voltage that can be set with the PSU (Volts).
        MAX_CURRENT: The maximum current that can be set with the PSU (Amperes).
    """
    MAX_VOLTAGE = 30
    MAX_CURRENT = 5


class Tenma72_2545(Tenma72_2550):
    """Instrument driver for the Tenma 72-2545, child of 72-2550.

    Attributes:
        MAX_VOLTAGE: The maximum voltage that can be set with the PSU (Volts).
        MAX_CURRENT: The maximum current that can be set with the PSU (Amperes).
    """
    MAX_VOLTAGE = 60
    MAX_CURRENT = 2


class Tenma72_2925(Tenma72_2550):
    """Instrument driver for the Tenma 72-2925, child of 72-2550.

    Attributes:
        MAX_VOLTAGE: The maximum voltage that can be set with the PSU (Volts).
        MAX_CURRENT: The maximum current that can be set with the PSU (Amperes).
    """
    MAX_VOLTAGE = 30
    MAX_CURRENT = 5


class Tenma72_2930(Tenma72_2550):
    """Instrument driver for the Tenma 72-2930, child of 72-2550.

    Attributes:
        MAX_VOLTAGE: The maximum voltage that can be set with the PSU (Volts).
        MAX_CURRENT: The maximum current that can be set with the PSU (Amperes).
    """
    MAX_VOLTAGE = 30
    MAX_CURRENT = 10


class Tenma72_2935(Tenma72_2550):
    """Instrument driver for the Tenma 72-2935, child of 72-2550.

    Attributes:
        MAX_VOLTAGE: The maximum voltage that can be set with the PSU (Volts).
        MAX_CURRENT: The maximum current that can be set with the PSU (Amperes).
    """
    MAX_VOLTAGE = 60
    MAX_CURRENT = 5


class Tenma72_2940(Tenma72_2550):
    """Instrument driver for the Tenma 72-2940, child of 72-2550.

    Attributes:
        MAX_VOLTAGE: The maximum voltage that can be set with the PSU (Volts).
        MAX_CURRENT: The maximum current that can be set with the PSU (Amperes).
    """
    MAX_VOLTAGE = 60
    MAX_CURRENT = 5


class Tenma72_10480(Tenma72_2550):
    """Instrument driver for the Tenma 72-10480, child of 72-2550.

    Attributes:
        MAX_VOLTAGE: The maximum voltage that can be set with the PSU (Volts).
        MAX_CURRENT: The maximum current that can be set with the PSU (Amperes).
    """
    MAX_VOLTAGE = 30
    MAX_CURRENT = 3


class Tenma72_13350(Tenma72_Base):
    """Instrument driver for the Tenma 72-13350. The driver is tested with this model, but manual is
    also for the model 72-13360.

    This driver can be used with (USB-to-)serial and UDP communications.

    Note that this model has extra functionalities related to the IP LAN Communication Protocol.
    Also note that this model does not work with channel number input, like 2550 model-based PSUs.
    """
    MAX_VOLTAGE = 30
    MAX_CURRENT = 30
    DEFAULT_BAUDRATE = 115200

    def __init__(self, context: QMI_Context, name: str, transport: str) -> None:
        super().__init__(context, name, transport)
        self.serial_eol = "\n"

    def _send(self, cmd: str) -> None:
        cmd = cmd + self.serial_eol
        return super()._send(cmd)

    @rpc_method
    def get_status(self) -> Dict[str, Any]:
        """Get the power supply status as a dictionary of status values.

        Dictionary composition is:
            "ChannelMode ":  "C.V" | "C.C"
            "OutputEnabled": True | False
            "V/C priority ": "Current priority" | "Voltage priority"
            "Beep":          True | False
            "Lock":          True | False

        Returns:
            Dictionary of status values.
        """
        self._send("STATUS?")
        status_bytes = self._transport.read(2, 2)  # Read response byte
        # 72-13350 sends two bytes back, the second being '\n'
        status = status_bytes[0]

        ch1mode = bool(status & 0x01)
        output_enabled = bool(status & 0x02)
        current_priority = status & 0x04
        beep = bool(status & 0x10)
        lock = bool(status & 0x20)

        return {
            "ChannelMode": TenmaChannelMode(ch1mode).name,
            "OutputEnabled": output_enabled,
            "V/C priority": "Current priority" if current_priority else "Voltage priority",
            "Beep": beep,
            "Lock": lock,
        }

    @rpc_method
    def enable_output(self, output: bool) -> None:
        command = f"OUT:{int(output)}"
        self._enable_output(output, command)

    @rpc_method
    def get_dhcp(self) -> int:
        """Use the IP LAN command to see if DHCP is enabled.

        Returns:
            dhcp: The current DHCP enabled state.
        """
        cmd = ":SYST:DHCP?"
        self._send(cmd)
        dhcp = self._read()
        return int(dhcp)

    @rpc_method
    def set_dhcp(self, dhcp: int) -> None:
        """Use the IP LAN command to set a DHCP enabled state.

        Parameters:
            dhcp: New DHCP state. 0 is disabled, 1 is enabled.
        """
        cmd = f":SYST:DHCP {dhcp}"
        self._send(cmd)

    @rpc_method
    def get_ip_address(self) -> str:
        """Use the IP LAN command to get the current IP address of the device.

        Returns:
            ip_address: The current IP address of the device.
        """
        cmd = ":SYST:IPAD?"
        self._send(cmd)
        ip_address = self._read()
        return ip_address

    @rpc_method
    def set_ip_address(self, ip_address) -> None:
        """Use the IP LAN command to set a new IP address for the device.

        Parameters:
            ip_address: A new static IP address for the device.
        """
        cmd = f":SYST:IPAD {ip_address}"
        self._send(cmd)

    @rpc_method
    def get_subnet_mask(self) -> str:
        """Use the IP LAN command to get the current subnet mask address of the device.

        Returns:
            subnet_mask: The current subnet mask address of the device.
        """
        cmd = ":SYST:SMASK?"
        self._send(cmd)
        subnet_mask = self._read()
        return subnet_mask

    @rpc_method
    def set_subnet_mask(self, subnet_mask) -> None:
        """Use the IP LAN command to set a new subnet mask address for the device.

        Parameters:
            subnet_mask: A new subnet mask address for the device.
        """
        cmd = f":SYST:SMASK {subnet_mask}"
        self._send(cmd)

    @rpc_method
    def get_gateway_address(self) -> str:
        """Use the IP LAN command to get the current gateway address of the device.

        Returns:
            gateway: The current static gateway address of the device.
        """
        cmd = ":SYST:GATE?"
        self._send(cmd)
        gateway = self._read()
        return gateway

    @rpc_method
    def set_gateway_address(self, gateway) -> None:
        """Use the IP LAN command to set a new gateway address for the device.

        Parameters:
            gateway: A new static gateway address for the device.
        """
        cmd = f":SYST:GATE {gateway}"
        self._send(cmd)

    @rpc_method
    def get_ip_port(self) -> int:
        """Use the IP LAN command to get the current IP port number of the device.

        Returns:
            ip_port: The current IP port number of the device.
        """
        cmd = ":SYST:PORT?"
        self._send(cmd)
        ip_port = self._read()
        return int(ip_port)

    @rpc_method
    def set_ip_port(self, ip_port: int) -> None:
        """Use the IP LAN command to set a new IP port number for the device.

        Parameters:
            ip_port: A new static IP port number for the device.
        """
        cmd = f":SYST:PORT {ip_port}"
        self._send(cmd)


class Tenma72_13360(Tenma72_13350):
    """Instrument driver for the Tenma 72-13360, child of 72-13350.

    Attributes:
        MAX_VOLTAGE: The maximum voltage that can be set with the PSU (Volts).
        MAX_CURRENT: The maximum current that can be set with the PSU (Amperes).
    """
    MAX_VOLTAGE = 60
    MAX_CURRENT = 15



"""Instrument driver for the Stanford Research Systems SIM922 Temperature Monitor """
from qmi.core.instrument import QMI_Instrument
from qmi.core.context import QMI_Context
from qmi.core.rpc import rpc_method
from qmi.instruments.stanford_research_systems.sim900 import Sim900
from qmi.core.exceptions import QMI_UsageException, QMI_InstrumentException


class SIM922(QMI_Instrument):
    """Instrument driver for the Stanford Research Systems SIM922 Temperature Monitor """

    def __init__(self, context: QMI_Context, name: str, sim900: Sim900, port: int) -> None:
        """Initialize driver.

        Arguments:
            name: Name for this instrument instance
            sim900: Instance of sim 900 module or RPC proxy of it, that hosts the amp-sim module.
            port: Port number in which the amp-sim module is located.
        """
        super().__init__(context, name)
        self._sim900 = sim900
        self._port = port
        self._channel_limits = (1, 4)

    @rpc_method
    def get_id(self) -> str:
        """ Query module identification string.

        Returns:
            Module identification string.
        """
        return self._sim900.ask_module(self._port, "*IDN?")

    @rpc_method
    def get_voltage(self, channel: int) -> float:
        """ Query the readout voltage on the specified channel.

        Returns:
            Readout voltage in millivolts.
        """
        self._is_valid(channel)
        voltage_string = self._sim900.ask_module(self._port, f"VOLT? {channel}")
        try:
            return float(voltage_string)
        except ValueError as error:
            raise QMI_InstrumentException(
                "Value returned by module should have been a string representation of a float, "
                f"instead received \"{voltage_string}\"."
            ) from error

    @rpc_method
    def get_temperature(self, channel: int) -> float:
        """ Query the temperature on the specified channel.

        Returns:
            Temperature in Kelvin.
        """
        self._is_valid(channel)
        temperature_string = self._sim900.ask_module(self._port, f"TVAL? {channel}")
        try:
            return float(temperature_string)
        except ValueError as error:
            raise QMI_InstrumentException(
                "Value returned by module should have been a string representation of a float, "
                f"instead received \"{temperature_string}\""
            ) from error

    @rpc_method
    def is_excited(self, channel: int) -> bool:
        """ Query the specified channel if excitation current is on.

        Arguments:
            channel: Channel to query
        Returns:
            True if excitation current is on, else False.
        """
        self._is_valid(channel)
        excitation_string = self._sim900.ask_module(self._port,f"EXON? {channel}").strip()
        if excitation_string in ("0", "OFF"):
            return False
        elif excitation_string in ("1", "ON"):
            return True
        else:
            raise QMI_InstrumentException(
                "Value returned by module should state whether excitation current is ON/1 or OFF/0, "
                f"instead received \"{excitation_string}\""
            )

    @rpc_method
    def set_excitation(self, channel: int, current_on: bool) -> None:
        """ Set current excitation to the specified channel to on/off.
        Arguments:
            channel: Channel to set excitation current.
            current_on: True if current on, else False.
        """
        self._is_valid(channel)
        if not isinstance(current_on, bool):
            raise QMI_UsageException(f"Expected parameter \"current_on\" to be of type bool, but is {current_on!r}")
        self._sim900.send_terminated_message(self._port, f"EXON {channel},{'ON' if current_on else 'OFF'}")

    def _is_valid(self, channel: int) -> None:
        if not self._channel_limits[0] <= channel <= self._channel_limits[1]:
            raise QMI_UsageException(
                f"Channel has to be between {self._channel_limits[0]} and "
                f"{self._channel_limits[1]}, but is {channel}.")

    def __repr__(self):
        return f"{self.__class__.__name__}("\
            f"context={self._context},"\
            f"name=\"{self._name}\","\
            f"sim900={self._sim900!r},"\
            f"port={self._port})"



"""
Instrument driver for the Rohde&Schwarz SGS100A RF Signal Generator
"""

import logging

from qmi.core.rpc import rpc_method
from qmi.instruments.rohde_schwarz.rs_base_signal_gen import RohdeSchwarz_Base

# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)


class RohdeSchwarz_SGS100A(RohdeSchwarz_Base):
    """Instrument driver for the Rohde&Schwarz SGS100A RF Signal Generator."""

    @rpc_method
    def get_external_reference_frequency(self) -> str:
        """Return the currently configured external reference input frequency.

        Possible values: "10MHZ", "100MHZ", "1000MHZ".
        """
        return self._get_external_reference_frequency()

    @rpc_method
    def set_external_reference_frequency(self, frequency: str) -> None:
        """Configure the external reference input frequency.

        Parameters:
            frequency: Desired frequency (accepted values: "10MHZ", "100MHZ", "1000MHZ"); see also
                       get_external_reference_frequency().
        """
        freq_options = ["10MHZ", "100MHZ", "1000MHZ"]
        self._set_external_reference_frequency(frequency, freq_options)

    @rpc_method
    def get_external_reference_bandwidth(self) -> str:
        """Return the current external reference synchronization bandwidth.

        Possible values:
          "NARR" - approximately 40 Hz synchronization bandwidth;
          "WIDE" - approximately 250 Hz synchronization bandwidth - suitable for high quality reference source.
        """
        self._check_calibrating()
        return self._scpi_protocol.ask(":ROSC:EXT:SBAN?").strip().upper()

    @rpc_method
    def set_external_reference_bandwidth(self, bandwidth: str) -> None:
        """Set the external reference synchronization bandwidth.

        Parameters:
            bandwidth:  desired bandwidth (accepted values: "NARR", "WIDE"); see also
                        get_external_reference_bandwidth().
        """
        options = ["NARR", "WIDE"]
        bandwidth = self._is_valid_param(bandwidth, options)

        self._check_calibrating()
        self._scpi_protocol.write(f":ROSC:EXT:SBAN {bandwidth}")
        self._check_error()

    @rpc_method
    def get_trigger_impedance(self) -> str:
        """Return current input impedance of the TRIG port.

        Possible values:
          "G50"  - 50 Ohm input impedance (default)
          "G10K" - 10 kOhm input impedance
        """
        self._check_calibrating()
        return self._scpi_protocol.ask(":PULM:TRIG:EXT:IMP?").strip().upper()

    @rpc_method
    def set_trigger_impedance(self, impedance: str) -> None:
        """Set the input impedance of the TRIG port.

        Parameters:
            impedance:  desired input impedance (accepted values: "G50", "G10K"); see also get_trigger_impedance().
        """
        options = ["G50", "G10K"]
        impedance = self._is_valid_param(impedance, options)

        self._check_calibrating()
        self._scpi_protocol.write(f":PULM:TRIG:EXT:IMP {impedance}")
        self._check_error()

    @rpc_method
    def start_calibration(self) -> None:
        super().start_calibration()
        self._scpi_protocol.write(":CAL:ALL?")

    @rpc_method
    def get_iq_correction_enabled(self) -> bool:
        """Return True if IQ modulation corrections are enabled, False if disabled."""
        self._check_calibrating()
        return self._ask_bool(":IQ:IMP:STAT?")

    @rpc_method
    def set_iq_correction_enabled(self, enable: bool) -> None:
        """Activate/deactivate the three correction values for the I/Q modulator.

        Parameters:
            enable: target enabled state.
        """
        self._check_calibrating()
        self._scpi_protocol.write(f":IQ:IMP:STAT {1 if enable else 0}")
        self._check_error()



"""Testcases of SIM922 temperature sensor."""
import unittest
from unittest.mock import call, MagicMock

import qmi
from qmi.instruments.stanford_research_systems import Srs_Sim900
from qmi.instruments.stanford_research_systems import Srs_Sim922
from qmi.core.exceptions import QMI_UsageException, QMI_InstrumentException


class TestSIM922(unittest.TestCase):
    """Testcase of SIM922 temperature sensor."""

    def setUp(self):
        qmi.start("test_siglent_sim922")
        self._sim900 = MagicMock(spec=Srs_Sim900)
        self._port = 1
        self._channel = 2
        self.sim922 = qmi.make_instrument("SIM922", Srs_Sim922, self._sim900, self._port)

    def tearDown(self):
        qmi.stop()

    def test_get_id(self):
        # arrange
        expected_id_string = "test_string"
        expected_calls = [
            call.ask_module(self._port, "*IDN?")
        ]
        self._sim900.ask_module.return_value = expected_id_string
        # act
        actual_id_string = self.sim922.get_id()
        # assert
        self.assertEqual(self._sim900.method_calls, expected_calls)
        self.assertEqual(actual_id_string, expected_id_string)

    def test_get_voltage(self):
        # arrange
        expected_voltage_string = "1.108054E+00"
        expected_voltage = 1.108054E+00
        expected_calls = [
            call.ask_module(self._port, f"VOLT? {self._channel}")
        ]
        self._sim900.ask_module.return_value = expected_voltage_string
        # act
        actual_voltage = self.sim922.get_voltage(self._channel)
        # assert
        self.assertEqual(self._sim900.method_calls, expected_calls)
        self.assertEqual(actual_voltage, expected_voltage)

    def test_get_voltage_channel_out_of_bounds(self):
        with self.assertRaises(QMI_UsageException):
            self.sim922.get_voltage(0)
        with self.assertRaises(QMI_UsageException):
            self.sim922.get_voltage(5)

    def test_get_voltage_wrong_voltage(self):
        # arrange
        self._sim900.ask_module.return_value = ""
        # act & assert
        with self.assertRaises(QMI_InstrumentException):
            self.sim922.get_voltage(1)

    def test_get_temperature(self):
        # arrange
        expected_temperature_string = "2.299000E+00"
        expected_temperature = 2.299000E+00
        expected_calls = [
            call.ask_module(self._port, f"TVAL? {self._channel}")
        ]
        self._sim900.ask_module.return_value = expected_temperature_string
        # act
        actual_temperature = self.sim922.get_temperature(self._channel)
        # assert
        self.assertEqual(self._sim900.method_calls, expected_calls)
        self.assertEqual(actual_temperature, expected_temperature)

    def test_get_temperature_channel_out_of_bounds(self):
        with self.assertRaises(QMI_UsageException):
            self.sim922.get_temperature(0)
        with self.assertRaises(QMI_UsageException):
            self.sim922.get_temperature(5)

    def test_get_temperature_wrong_temperature(self):
        # arrange
        self._sim900.ask_module.return_value = ""
        # act & assert
        with self.assertRaises(QMI_InstrumentException):
            self.sim922.get_temperature(self._channel)

    def test_is_excited_(self):
        expected_excitation_list = [('1', True), ('ON', True), ('0', False), ('OFF', False)]
        for expected_excitation in expected_excitation_list:
            self._sim900.ask_module.return_value = expected_excitation[0]
            with self.subTest(i=expected_excitation[0]):
                self.assertEqual(self.sim922.is_excited(self._channel), expected_excitation[1])

    def test_set_excitation(self):
        expected_call_list = [
            (True, [call.send_terminated_message(self._port, f"EXON {self._channel},ON")]),
            (False, [call.send_terminated_message(self._port, f"EXON {self._channel},OFF")])
        ]
        for expected_call in expected_call_list:
            self._sim900.reset_mock()
            self.sim922.set_excitation(self._channel, current_on=expected_call[0])
            with self.subTest(i=expected_call[0]):
                self.assertEqual(self._sim900.method_calls, expected_call[1])

    def test_set_excitation_not_bool(self):
        with self.assertRaises(QMI_UsageException):
            self.sim922.set_excitation(self._channel, current_on=None)


if __name__ == '___main___':
    unittest.main()



import ctypes
import time
import unittest
from fractions import Fraction
from unittest.mock import patch

from qmi.core.exceptions import QMI_InstrumentException, QMI_InvalidOperationException
from qmi.instruments.picoquant.support._mhlib_function_signatures import _mhlib_function_signatures
from qmi.instruments.picoquant import PicoQuant_MultiHarp150

from tests.patcher import PatcherQmiContext as QMI_Context


class MultiHarpOldLibrayTestCase(unittest.TestCase):

    def test_initialize_excepts_valid_mode_but_too_old_lib(self):
        patcher = patch('ctypes.cdll.LoadLibrary', spec=ctypes.CDLL)
        self._library_mock = patcher.start().return_value

        function_names, _, _ = zip(*_mhlib_function_signatures)
        self._library_mock.mock_add_spec(function_names, spec_set=True)

        self.addCleanup(patcher.stop)

        self._multiharp: PicoQuant_MultiHarp150 = PicoQuant_MultiHarp150(
            QMI_Context("Test_multiharp"), 'multiharp', '1111111'
        )

        self._library_mock.MH_GetLibraryVersion.return_value = 0
        self._library_mock.MH_OpenDevice.return_value = 0

        string_buffer = ctypes.create_string_buffer(b'2.222')
        string_buffer2 = ctypes.create_string_buffer(b'1111111')
        retvals = (string_buffer, string_buffer2)

        with patch('sys.platform', 'linux1'), patch('ctypes.create_string_buffer', side_effect=retvals):
            self._multiharp.open()

        self._library_mock.MH_Initialize.return_value = 0
        with self.assertRaises(ValueError):
            self._multiharp.initialize('T3', 'INTERNAL')

        self._library_mock.MH_CloseDevice.return_value = 0
        self._multiharp.close()


class MultiharpMethodsTestCase(unittest.TestCase):

    def setUp(self) -> None:
        patcher = patch('ctypes.cdll.LoadLibrary', spec=ctypes.CDLL)
        self._library_mock = patcher.start().return_value

        function_names, _, _ = zip(*_mhlib_function_signatures)
        self._library_mock.mock_add_spec(function_names, spec_set=True)

        self.addCleanup(patcher.stop)

        self._multiharp: PicoQuant_MultiHarp150 = PicoQuant_MultiHarp150(
            QMI_Context("test_multiharp"), 'multiharp', '1111111'
        )

        self._library_mock.MH_GetLibraryVersion.return_value = 0
        self._library_mock.MH_OpenDevice.return_value = 0

        string_buffer = ctypes.create_string_buffer(b'1111111')

        with patch('sys.platform', 'linux1'), patch('ctypes.create_string_buffer', return_value=string_buffer):
            self._multiharp.open()

    def tearDown(self) -> None:
        self._library_mock.MH_CloseDevice.return_value = 0
        self._multiharp.close()

    def test_open_mh(self):
        """Test MultiHarp open where regular open fails and we need to get the SN with a query."""
        self._library_mock.MH_CloseDevice.return_value = 0
        self._multiharp.close()

        self._library_mock.MH_GetLibraryVersion.return_value = 0
        self._library_mock.MH_OpenDevice.return_value = 0
        self._library_mock.MH_Initialize.return_value = 0
        self._library_mock.MH_GetSerialNumber.return_value = 0

        string_buffer_lib = ctypes.create_string_buffer(8)
        string_buffer_lib.value = b'1.23.4'
        string_buffer_sn1 = ctypes.create_string_buffer(8)
        string_buffer_sn1.value = b''
        string_buffer_sn2 = ctypes.create_string_buffer(8)
        string_buffer_sn2.value = b'1111111'

        with patch('sys.platform', 'linux1'):
            with patch('ctypes.create_string_buffer', side_effect=[string_buffer_lib, string_buffer_sn1, string_buffer_sn2]):
                self._multiharp.open()

            self._library_mock.MH_Initialize.assert_called_once()
            self._library_mock.MH_GetSerialNumber.assert_called_once()

    def test_initialize(self):

        self._library_mock.MH_Initialize.return_value = 0

        self._multiharp.initialize('T2', 'INTERNAL')

    def test_initialize_excepts_measurement_running(self):
        self._library_mock.MH_Initialize.return_value = 0
        self._library_mock.MH_StartMeas.return_value = 0
        self._multiharp.start_measurement(1000)
        with self.assertRaises(QMI_InvalidOperationException):
            self._multiharp.initialize('T2', 'INTERNAL')

    def test_initialize_excepts_invalid_mode(self):
        self._library_mock.MH_Initialize.return_value = 0
        with self.assertRaises(ValueError):
            self._multiharp.initialize('T4', 'INTERNAL')

    def test_initialise_raises_libraryError(self):

        self._library_mock.MH_Initialize.return_value = -1
        self._library_mock.MH_GetErrorString.return_value = 0

        error_string = 'error string'
        ctype_err_string = ctypes.create_string_buffer(error_string.encode('ASCII'), 40)

        with self.assertRaises(QMI_InstrumentException), \
                patch('ctypes.create_string_buffer', return_value=ctype_err_string):

            self._multiharp.initialize('T2', 'INTERNAL')

    def test_get_hardware_info(self):

        self._library_mock.MH_GetHardwareInfo.return_value = 0
        expected_hardware_info = 'abcdefgh', '12345678', '23456789'

        model_string_buffer = ctypes.create_string_buffer(expected_hardware_info[0].encode('ASCII'), 24)
        partno = ctypes.create_string_buffer(expected_hardware_info[1].encode('ASCII'), 8)
        version = ctypes.create_string_buffer(expected_hardware_info[2].encode('ASCII'), 8)

        with patch('ctypes.create_string_buffer', side_effect=[model_string_buffer, partno, version]):
            hardware_info = self._multiharp.get_hardware_info()

        self.assertTupleEqual(hardware_info, expected_hardware_info)

    def test_get_features(self):

        self._library_mock.MH_GetFeatures.return_value = 0

        bitpattern = 0x0009

        with patch('ctypes.c_int', return_value=ctypes.c_int(bitpattern)):
            features = self._multiharp.get_features()

        self.assertIn('DLL', features)
        self.assertIn('LOWRES', features)

    def test_module_info(self):
        self._library_mock.MH_GetNumOfModules.return_value = 0
        self._library_mock.MH_GetModuleInfo.return_value = 0

        expected_module_info = [(1, 2), (3, 4)]

        values = [2, 1, 2, 3, 4]
        patcher = patch('ctypes.c_int', side_effect=[ctypes.c_int(value) for value in values])

        with patcher:
            module_info = self._multiharp.get_module_info()

        for actual, expected in zip(module_info, expected_module_info):
            self.assertTupleEqual(actual, expected)

    def test_get_debug_info(self):
        self._library_mock.MH_GetDebugInfo.return_value = 0
        expected_debug_info = 'foo bar baz'

        string_buffer = ctypes.create_string_buffer(expected_debug_info.encode('ASCII'), 65536)

        with patch('ctypes.create_string_buffer', return_value=string_buffer):
            debug_info = self._multiharp.get_debug_info()

        self.assertEqual(debug_info, expected_debug_info)

    def test_set_sync_channel_offset(self):

        self._library_mock.MH_SetSyncChannelOffset.return_value = 0
        self._multiharp.set_sync_channel_offset(99999)
        self._library_mock.MH_SetSyncChannelOffset.assert_called_once_with(0, 99999)

    def test_set_sync_dead_time(self):

        self._library_mock.MH_SetSyncDeadTime.return_value = 0
        self._multiharp.set_sync_dead_time(0, 160000)
        self._library_mock.MH_SetSyncDeadTime.assert_called_once_with(0, 0, 160000)

    def test_set_input_edge_trigger(self):

        self._library_mock.MH_SetInputEdgeTrg.return_value = 0
        self._multiharp.set_input_edge_trigger(1, 1000, 'RISING')
        self._library_mock.MH_SetInputEdgeTrg.assert_called_once_with(0, 1, 1000, 1)

    def test_set_input_channel_dead_time(self):

        self._library_mock.MH_SetInputDeadTime.return_value = 0
        self._multiharp.set_input_channel_dead_time(1, 0, 800)
        self._library_mock.MH_SetInputDeadTime.assert_called_once_with(0, 1, 0, 800)

    def test_set_measurement_control(self):

        self._library_mock.MH_SetMeasControl.return_value = 0
        self._multiharp.set_measurement_control('SINGLESHOT_CTC', 'RISING', 'FALLING')
        self._library_mock.MH_SetMeasControl.assert_called_once_with(0, 0, 1, 0)

    def test_set_trigger_output(self):

        self._library_mock.MH_SetTriggerOutput.return_value = 0
        self._multiharp.set_trigger_output(16777215)
        self._library_mock.MH_SetTriggerOutput.assert_called_once_with(0, 16777215)

    def test_get_all_count_rates(self):

        self._library_mock.MH_GetNumOfInputChannels.return_value = 0
        self._library_mock.MH_GetAllCountRates.return_value = 0

        num_channels = 4
        expected_sync_rate = 10
        expected_input_rates = (11, 12, 13, 14)

        values = (num_channels, expected_sync_rate, *expected_input_rates)
        with patch('ctypes.c_int', side_effect=[ctypes.c_int(value) for value in values]):
            sync_rate, input_rates = self._multiharp.get_all_count_rates()

        self.assertEqual(sync_rate, expected_sync_rate)

    def test_get_flags(self):

        self._library_mock.MH_GetFlags.return_value = 0

        flag_bitset = 0b01

        with patch('ctypes.c_int', return_value=ctypes.c_int(flag_bitset)):
            flags = self._multiharp.get_flags()

        self.assertIn("OVERFLOW", flags)

    def test_get_warnings(self):

        self._library_mock.MH_GetWarnings.return_value = 0

        flag_bitset = 0b01

        with patch('ctypes.c_int', return_value=ctypes.c_int(flag_bitset)):
            flags = self._multiharp.get_warnings()

        self.assertIn("SYNC_RATE_ZERO", flags)

    def test_set_sync_edge_trigger(self):

        self._library_mock.MH_SetSyncEdgeTrg.return_value = 0
        self._multiharp.set_sync_edge_trigger(-1200, 'RISING')
        self._library_mock.MH_SetSyncEdgeTrg.assert_called_once_with(0, -1200, 1)

    def test_get_start_time(self):

        self._library_mock.MH_GetStartTime.return_value = 0

        with patch('ctypes.c_uint', return_value=ctypes.c_uint(1)):
            start_time = self._multiharp.get_start_time()

        self.assertEqual(start_time, Fraction((1 << 64) | (1 << 32) | 1, 1000000000000))

    def test_get_start_time2(self):

        self._library_mock.MH_GetStartTime.return_value = 0

        picoseconds = (10 << 64) | (9 << 32) | 8

        expected_start_time = Fraction(picoseconds, 1000000000000)

        values = (10, 9, 8)
        with patch('ctypes.c_uint', side_effect=[ctypes.c_uint(value) for value in values]):
            start_time = self._multiharp.get_start_time()

        self.assertEqual(start_time, expected_start_time)

    def test_get_histogram(self):
        """Test on get_histogram function without 'clear' input variable"""
        self._library_mock.MH_GetHistogram.return_value = 0

        self._multiharp.get_histogram(1, 0)
        self._library_mock.MH_GetHistogram.assert_called()
        self.assertEqual(len(self._library_mock.MH_GetHistogram.call_args_list[0][0]), 4)

    def test_get_all_histograms(self):
        """Test on get_all_histograms function."""
        self._library_mock.MH_GetNumOfInputChannels.return_value = 0
        self._library_mock.MH_GetAllHistograms.return_value = 0
        no_of_channels = 2
        # Mock with no_of_channels that will have 65536 data entry points
        with patch('ctypes.c_int', return_value=ctypes.c_int(no_of_channels)):
            retval = self._multiharp.get_all_histograms()
            self.assertTupleEqual(retval.shape, (no_of_channels, 65536))
            self._library_mock.MH_GetAllHistograms.assert_called()
            self.assertEqual(len(self._library_mock.MH_GetAllHistograms.call_args_list[0][0]), 2)

    def test_read_fifo(self):
        self._library_mock.MH_ReadFiFo.return_value = 0
        self._library_mock.MH_StartMeas.return_value = 0
        self._library_mock.MH_StopMeas.return_value = 0
        self._library_mock.MH_Initialize.return_value = 0

        self._multiharp.initialize('T2', 'INTERNAL')
        self._multiharp.start_measurement(1)
        time.sleep(0.01)
        self._multiharp.stop_measurement()

        self._library_mock.MH_ReadFiFo.assert_called()
        self.assertEqual(len(self._library_mock.MH_ReadFiFo.call_args_list[0][0]), 3)


if __name__ == '__main__':
    unittest.main()



"""A Python wrapper module to wrap around manufacturer provided library header file 'hhdefin.h'. The header file used
is from software version 3.0.0.4.
Please find the header file in the dedicated software package for the HydraHarp instrument at
https://www.picoquant.com/dl_software/HydraHarp400/HydraHarp400_SW_and_DLL_v3_0_0_4.zip"""
# This is generated code, do not edit by hand!
import ctypes


_hhlib_function_signatures = [
        ('HH_GetLibraryVersion', ctypes.c_int, [('vers', ctypes.POINTER(ctypes.c_char))]),
        ('HH_GetErrorString', ctypes.c_int, [('errstring', ctypes.POINTER(ctypes.c_char)), ('errcode', ctypes.c_int)]),
        ('HH_OpenDevice', ctypes.c_int, [('devidx', ctypes.c_int), ('serial', ctypes.POINTER(ctypes.c_char))]),
        ('HH_CloseDevice', ctypes.c_int, [('devidx', ctypes.c_int)]),
        ('HH_Initialize', ctypes.c_int, [('devidx', ctypes.c_int), ('mode', ctypes.c_int), ('refsource', ctypes.c_int)]),
        ('HH_GetHardwareInfo', ctypes.c_int, [('devidx', ctypes.c_int), ('model', ctypes.POINTER(ctypes.c_char)), ('partno', ctypes.POINTER(ctypes.c_char)), ('version', ctypes.POINTER(ctypes.c_char))]),
        ('HH_GetSerialNumber', ctypes.c_int, [('devidx', ctypes.c_int), ('serial', ctypes.POINTER(ctypes.c_char))]),
        ('HH_GetFeatures', ctypes.c_int, [('devidx', ctypes.c_int), ('features', ctypes.POINTER(ctypes.c_int))]),
        ('HH_GetBaseResolution', ctypes.c_int, [('devidx', ctypes.c_int), ('resolution', ctypes.POINTER(ctypes.c_double)), ('binsteps', ctypes.POINTER(ctypes.c_int))]),
        ('HH_GetHardwareDebugInfo', ctypes.c_int, [('devidx', ctypes.c_int), ('debuginfo', ctypes.POINTER(ctypes.c_char))]),
        ('HH_GetNumOfInputChannels', ctypes.c_int, [('devidx', ctypes.c_int), ('nchannels', ctypes.POINTER(ctypes.c_int))]),
        ('HH_GetNumOfModules', ctypes.c_int, [('devidx', ctypes.c_int), ('nummod', ctypes.POINTER(ctypes.c_int))]),
        ('HH_GetModuleInfo', ctypes.c_int, [('devidx', ctypes.c_int), ('modidx', ctypes.c_int), ('modelcode', ctypes.POINTER(ctypes.c_int)), ('versioncode', ctypes.POINTER(ctypes.c_int))]),
        ('HH_GetModuleIndex', ctypes.c_int, [('devidx', ctypes.c_int), ('channel', ctypes.c_int), ('modidx', ctypes.POINTER(ctypes.c_int))]),
        ('HH_Calibrate', ctypes.c_int, [('devidx', ctypes.c_int)]),
        ('HH_SetSyncDiv', ctypes.c_int, [('devidx', ctypes.c_int), ('div', ctypes.c_int)]),
        ('HH_SetSyncCFD', ctypes.c_int, [('devidx', ctypes.c_int), ('level', ctypes.c_int), ('zc', ctypes.c_int)]),
        ('HH_SetSyncChannelOffset', ctypes.c_int, [('devidx', ctypes.c_int), ('value', ctypes.c_int)]),
        ('HH_SetInputCFD', ctypes.c_int, [('devidx', ctypes.c_int), ('channel', ctypes.c_int), ('level', ctypes.c_int), ('zc', ctypes.c_int)]),
        ('HH_SetInputChannelOffset', ctypes.c_int, [('devidx', ctypes.c_int), ('channel', ctypes.c_int), ('value', ctypes.c_int)]),
        ('HH_SetInputChannelEnable', ctypes.c_int, [('devidx', ctypes.c_int), ('channel', ctypes.c_int), ('enable', ctypes.c_int)]),
        ('HH_SetStopOverflow', ctypes.c_int, [('devidx', ctypes.c_int), ('stop_ovfl', ctypes.c_int), ('stopcount', ctypes.c_uint)]),
        ('HH_SetBinning', ctypes.c_int, [('devidx', ctypes.c_int), ('binning', ctypes.c_int)]),
        ('HH_SetOffset', ctypes.c_int, [('devidx', ctypes.c_int), ('offset', ctypes.c_int)]),
        ('HH_SetHistoLen', ctypes.c_int, [('devidx', ctypes.c_int), ('lencode', ctypes.c_int), ('actuallen', ctypes.POINTER(ctypes.c_int))]),
        ('HH_SetMeasControl', ctypes.c_int, [('devidx', ctypes.c_int), ('control', ctypes.c_int), ('startedge', ctypes.c_int), ('stopedge', ctypes.c_int)]),
        ('HH_ClearHistMem', ctypes.c_int, [('devidx', ctypes.c_int)]),
        ('HH_StartMeas', ctypes.c_int, [('devidx', ctypes.c_int), ('tacq', ctypes.c_int)]),
        ('HH_StopMeas', ctypes.c_int, [('devidx', ctypes.c_int)]),
        ('HH_CTCStatus', ctypes.c_int, [('devidx', ctypes.c_int), ('ctcstatus', ctypes.POINTER(ctypes.c_int))]),
        ('HH_GetHistogram', ctypes.c_int, [('devidx', ctypes.c_int), ('chcount', ctypes.POINTER(ctypes.c_uint)), ('channel', ctypes.c_int), ('clear', ctypes.c_int)]),
        ('HH_GetResolution', ctypes.c_int, [('devidx', ctypes.c_int), ('resolution', ctypes.POINTER(ctypes.c_double))]),
        ('HH_GetSyncPeriod', ctypes.c_int, [('devidx', ctypes.c_int), ('period', ctypes.POINTER(ctypes.c_double))]),
        ('HH_GetSyncRate', ctypes.c_int, [('devidx', ctypes.c_int), ('syncrate', ctypes.POINTER(ctypes.c_int))]),
        ('HH_GetCountRate', ctypes.c_int, [('devidx', ctypes.c_int), ('channel', ctypes.c_int), ('cntrate', ctypes.POINTER(ctypes.c_int))]),
        ('HH_GetFlags', ctypes.c_int, [('devidx', ctypes.c_int), ('flags', ctypes.POINTER(ctypes.c_int))]),
        ('HH_GetElapsedMeasTime', ctypes.c_int, [('devidx', ctypes.c_int), ('elapsed', ctypes.POINTER(ctypes.c_double))]),
        ('HH_GetWarnings', ctypes.c_int, [('devidx', ctypes.c_int), ('warnings', ctypes.POINTER(ctypes.c_int))]),
        ('HH_GetWarningsText', ctypes.c_int, [('devidx', ctypes.c_int), ('text', ctypes.POINTER(ctypes.c_char)), ('warnings', ctypes.c_int)]),
        ('HH_SetMarkerHoldoffTime', ctypes.c_int, [('devidx', ctypes.c_int), ('holdofftime', ctypes.c_int)]),
        ('HH_SetMarkerEdges', ctypes.c_int, [('devidx', ctypes.c_int), ('me1', ctypes.c_int), ('me2', ctypes.c_int), ('me3', ctypes.c_int), ('me4', ctypes.c_int)]),
        ('HH_SetMarkerEnable', ctypes.c_int, [('devidx', ctypes.c_int), ('en1', ctypes.c_int), ('en2', ctypes.c_int), ('en3', ctypes.c_int), ('en4', ctypes.c_int)]),
        ('HH_ReadFiFo', ctypes.c_int, [('devidx', ctypes.c_int), ('buffer', ctypes.POINTER(ctypes.c_uint)), ('count', ctypes.c_int), ('nactual', ctypes.POINTER(ctypes.c_int))]),
        ('HH_GetContModeBlock', ctypes.c_int, [('devidx', ctypes.c_int), ('buffer', ctypes.c_void_p), ('nbytesreceived', ctypes.POINTER(ctypes.c_int))])
    ]



"""Test for the Ozoptics DD100MC driver."""

from unittest import TestCase
from unittest.mock import MagicMock
from unittest.mock import patch
from unittest.mock import call

from typing import cast

from dataclasses import dataclass


from qmi.core.exceptions import QMI_InstrumentException
from qmi.core.transport import QMI_Transport
from qmi.instruments.ozoptics.dd100mc import OZOptics_DD100MC


@dataclass
class TestMeta:
    """Test meta data."""

    transport: MagicMock
    transport_str: MagicMock
    name: MagicMock
    super: MagicMock
    instr: OZOptics_DD100MC


class TestOZOptics_DD100MC(TestCase):
    """Testcase for the OZOptics_DD100MC class."""

    def setUp(self):
        mock_transport = MagicMock(spec=QMI_Transport)
        mock_name = MagicMock()
        mock_transport_str = MagicMock()
        mock_super = MagicMock()

        with patch(
            "qmi.instruments.ozoptics.dd100mc.create_transport",
            return_value=mock_transport,
        ):
            instr = OZOptics_DD100MC(MagicMock(), mock_name, mock_transport_str)

        self._patcher_super = patch(
            "qmi.instruments.ozoptics.dd100mc.super", mock_super
        )
        self._patcher_super.start()

        self._meta = TestMeta(
            transport=mock_transport,
            instr=cast(OZOptics_DD100MC, instr),
            transport_str=mock_transport_str,
            name=mock_name,
            super=mock_super,
        )

    def tearDown(self):
        self._meta = None
        self._patcher_super.stop()

    def test_init(self):
        """OZOptics_DD100MC.__init__(), happy flow."""
        self.assertEqual(self._meta.instr._transport, self._meta.transport)

    def test_open(self):
        """OZOptics_DD100MC.open(), happy flow."""
        self._meta.instr.open()
        self._meta.transport.open.assert_called_once_with()
        self._meta.transport.discard_read.assert_called_once_with()
        self._meta.super().open.assert_called_once_with()

    def test_close(self):
        """OZOptics_DD100MC.close(), happy flow."""
        self._meta.instr.close()
        self._meta.transport.close.assert_called_once_with()
        self._meta.super().close.assert_called_once_with()

    def test_read_response(self):
        """OZOptics_DD100MC._read_response(), happy flow."""
        mock_timeout = MagicMock()
        mock_resp = [i.encode("ascii") for i in ["RESPONSE\r\n", "Done\r\n", "\r\n"]]

        self._meta.instr._check_is_open = MagicMock()
        self._meta.transport.read_until = MagicMock(side_effect=mock_resp)

        rt_val = self._meta.instr._read_response(mock_timeout)

        self.assertEqual(rt_val[0], "RESPONSE")
        self._meta.instr._check_is_open.assert_called_once_with()
        self._meta.transport.read_until.assert_has_calls(
            [
                call(b"\r\n", mock_timeout),
            ]
            * 3
        )

    def test_execute_command(self):
        """OZOptics_DD100MC._execute_command(), happy flow."""
        mock_cmd = "CD"
        mock_timeout = MagicMock()

        self._meta.instr._check_is_open = MagicMock()
        self._meta.instr._read_response = MagicMock()

        rt_val = self._meta.instr._execute_command(mock_cmd, mock_timeout)
        self._meta.transport.write.assert_called_once_with(
            (mock_cmd + "\r").encode("ascii")
        )
        self._meta.instr._check_is_open.assert_called_once_with()
        self._meta.instr._read_response.assert_called_once_with(mock_timeout)
        self.assertEqual(rt_val, self._meta.instr._read_response())

    def test_get_configuration_display(self):
        """OZOptics_DD100MC.get_configuration_display(), happy flow."""

        self._meta.instr._execute_command = MagicMock()
        rt_val = self._meta.instr.get_configuration_display()

        self._meta.instr._execute_command.assert_called_once_with(
            "CD", self._meta.instr.RESPONSE_TIMEOUT
        )
        self.assertEqual(rt_val, self._meta.instr._execute_command())

    def test_home(self):
        """OZOptics_DD100MC.home(), happy flow."""

        self._meta.instr._execute_command = MagicMock()
        self._meta.instr.home()

        self._meta.instr._execute_command.assert_called_once_with(
            "H", self._meta.instr.RESPONSE_TIMEOUT_HOME_COMMAND
        )

    def test_get_position(self):
        """OZOptics_DD100MC.get_position(), happy flow."""
        mock_steps = 10000
        mock_atten = 0.0001
        mock_resp = [f"DPos:{mock_steps}", f"ATTEN:{mock_atten}"]
        self._meta.instr._execute_command = MagicMock(return_value=mock_resp)
        rt_val = self._meta.instr.get_position()

        self._meta.instr._execute_command.assert_called_once_with(
            "D", self._meta.instr.RESPONSE_TIMEOUT
        )
        self.assertEqual(rt_val.steps, mock_steps)
        self.assertEqual(rt_val.attenuation, mock_atten)

    def test_get_position_invalid_response(self):
        """OZOptics_DD100MC.get_position(), invalid response handling."""

        def test_response(response):
            self._meta.instr._execute_command = MagicMock(return_value=response)
            with self.assertRaises(QMI_InstrumentException):
                self._meta.instr.get_position()

        test_response("")
        test_response(["DPos:10000", ""])
        test_response(["", "ATTEN:0.0001"])

    def test_set_attenuation(self):
        """OZOptics_DD100MC.set_attenuation(), happy flow."""

        mock_value = 0.01
        mock_pos = 200
        mock_resp = [f"Pos:{mock_pos}"]
        self._meta.instr._execute_command = MagicMock(return_value=mock_resp)

        rt_val = self._meta.instr.set_attenuation(mock_value)

        self._meta.instr._execute_command.assert_called_once_with(
            f"A{mock_value:.2f}", self._meta.instr.RESPONSE_TIMEOUT
        )
        self.assertEqual(rt_val, mock_pos)

    def test_set_attenuation_invalid_response(self):
        """OZOptics_DD100MC.set_attenuation(), invalid response handling."""

        def test_response(response):
            self._meta.instr._execute_command = MagicMock(return_value=response)
            with self.assertRaises(QMI_InstrumentException):
                self._meta.instr.set_attenuation(0.01)

        test_response("")
        test_response([""])
